#!/usr/local/bin/perl -w



=head1 NAME



wsdshell.pl



=head1 SYNOPSIS



This program is a wrapper script to run supervised Word Sense Disambiguation experiments using SenseTools, Ngram Statistics Package and Weka, on any WSD data that is in SENSEVAL2 format.



=head1 DESCRIPTION



Running supervised word sense disambiguation experiments using the Weka Data Mining Suite requires that the data is represented in a format that Weka can interpret (the ARFF format). Prior to that, the task of feature identification and extraction must be performed. Ngram Statistics Package provides us with programs for feature identification and extraction, while SenseTools has programs for converting these features into the ARFF format required by Weka.



wsdshell.pl puts all these programs together, adding some "ease of use" features. It accepts command line arguments that are common to two or more programs that it is going to invoke and provides the flexibility to pass additional program specific arguments. For a detailed list of options, see 'wsdshell.pl --help'.



Since the program is just a wrapper which calls other programs in sequence, the user can specify if the program should skip some steps and directly jump to any particular step and continue execution from that point - e.g. directly jumping to step 6 below. This helps if for example we simply want to run a different classifier on the same set of extracted features.



The exact sequence that is executed is as follows:



1. preprocess.pl (SenseTools)



This is executed to identify tokens and non-tokens in the input SENSEVAL2 data file. Two output files with ".count" (for count.pl) and ".xml" (for xml2arff.pl)  extensions are created as a result.



2. count.pl (NSP)



This is run on the .count file produced by preprocess.pl to identify Ngram features and create an output file readable by NSP statistic.pl program. One output file - ".cnt" (for statistic.pl) is created as output.



3. statistic.pl (NSP)



This is run to selectively identify features of interest out of the features extracted by count.pl program. One ".stat" file is created as output. This step is skipped if the NGRAM value is 1.



4. nsp2regex.pl (SenseTools)



This program converts the features of interest identified by statistic.pl into a set of regular experessions that can be used to convert the data to ARFF format. One ".regex" file is created as output. When statistic.pl is skipped, this program operated directly on the output of count.pl program.



5. xml2arff.pl (SenseTools)



It takes the ".xml" file created by preprocess.pl and the ".regex" file created by nsp2regex.pl and creates a ".arff" file as output, in sparse ARFF format.



6. Weka classifier



This is an invokation of the specified Weka classifier on the ".arff" file created by xml2arff.pl program. The output is either at STDOUT if no output file name is specified or the specified output file contains the Weka output.





=head1 USAGE



See 'wsdshell.pl --help' for a list of options and their usage.



=head1 AUTHOR



Mahesh Joshi <joshi031@d.umn.edu>



Ted Pedersen <tpederse@d.umn.edu>



=head1 BUGS



=head1 SEE ALSO



Senseval-2  :  http://www.sle.sharp.co.uk/senseval2/



SenseTools  :  http://www.d.umn.edu/~tpederse/sensetools.html



Ngram Statistics Package : http://www.d.umn.edu/~tpederse/nsp.html



The WEKA Data Mining Suite: http://www.cs.waikato.ac.nz/ml/weka



wsdshelldriver.pl



runtests.pl



wsdsummary.pl



=head1 COPYRIGHT



Copyright (C) 2005, Ted Pedersen and Mahesh Joshi



This program is free software; you can redistribute it and/or

modify it under the terms of the GNU General Public License

as published by the Free Software Foundation; either version 2

of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,

but WITHOUT ANY WARRANTY; without even the implied warranty of

MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

GNU General Public License for more details.



You should have received a copy of the GNU General Public License

along with this program; if not, write to the Free Software

Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.



=cut



##############################################################################

############################## CHANGE LOG ####################################

##############################################################################



# 1. Date: 01/18/2005

#    Modified by: Mahesh Joshi

#    Comments: First version of script ready, without POD documentation



# 2. Date: 01/21/2005

#    Modified by: Mahesh Joshi

#    Comments: Added POD documentation



##############################################################################

########################### END OF CHANGE LOG ################################

##############################################################################





use strict;



use Getopt::Long;



# Variable declarations



# Binary variables that determine the verbosity and the trace output of the program

my $verbose = 0;

my $trace = 0;



# Checks if user has requested help at command line

my $help = 0;



# Variables for various file names

my $tokenfile;

my $nontokenfile;

my $stopfile;



my $inputsval2file;

my $wekaoutputfile;



# File handle for input

my $hin;



# Variables controlling the Weka invokation

my $wekaclassifier = "weka.classifiers.functions.SMO"; # Default

my $wekaopts = "";



# Variables to hold options of other programs

my $prepopts;

my $countopts;

my $statopts;



my $ngrams = 2;



# statistic.pl related options

my $statlibrary = "ll.pm"; # Default



# Variable that decides what step should the program jump to directly

my $step = 1; # Start from the beginning by default



#######################

# SUB-ROUTINE SECTION #

#######################



#############

# showUsage #

#############



# This function prints out the usage of the program to STDOUT in a terse format

sub showUsage() {



    # Get the name of the program by which it is running currently

    my @tokens = split /\//, $0;

    my $filename = $tokens[@tokens-1];



    print "Usage: $filename [OPTIONS] INPUT_SENSEVAL2_FILE [WEKA_OUTPUT_FILE]\n";

    print "Type \"$filename --help\" for a list of options or \"perldoc $filename\" for details.\n";

}



############

# showHelp #

############



# This function prints out an intermediate level of usage information to STDOUT

# For further details, user is expected to read the perldoc for the program.

sub showHelp() {



    # Get the name of the program by which it is running currently

    my @tokens = split /\//, $0;

    my $filename = $tokens[@tokens-1];



    print "\n$filename: Driver program to run Word Sense Disambiguation (WSD)\n";

    print "experiments on a data file in SENSEVAL2 format.\n\n";



    print "Usage: $filename [OPTIONS] INPUT_SENSEVAL2_FILE [WEKA_OUPUT_FILE]\n\n";



    print "Specifying the input file in SENSEVAL2 format is mandatory.\n\n";



    print "Valid options are:\n\n";



    print "--token FILE     Specifies the name of the file containing regular\n";

    print "                 expressions (one per line) to identify tokens. The\n";

    print "                 file is passed as an argument to preprocess.pl, \n";

    print "                 count.pl and nsp2regex.pl programs.\n";

    print "                 e.g. params/tokens file in wsdshell directory.\n\n";



    print "--nontoken FILE  Specifies the name of the file containing regular\n";

    print "                 experessions (one per line) to identify non tokens.\n";

    print "                 The file is passed as an argument to preprocess.pl \n";

    print "                 and count.pl programs.\n";



    print "                 e.g. params/nontokens file in wsdshell directory.\n\n";



    print "--stoplist FILE  Specifies the name of the file containing the list of\n";

    print "                 stop words to be ignored by the count.pl program. The\n";

    print "                 words should be specified as regular expressions with\n";

    print "                 an AND or OR mode. Refer to documentation of count.pl\n";

    print "                 for details.\n";

    print "                 e.g. params/stoplist file in wsdshell directory.\n\n";



    print "--statlib LIBRARY\n";

    print "                 Specifies the statistic measure that should be applied\n";

    print "                 by the statistic.pl program. The default is log \n";

    print "                 likelihood measure (ll.pm).\n\n";



    print "--ngram N        Specifies the number of words to be considered in an \n";

    print "                 ngram by count.pl and statistic.pl programs.\n\n";



    print "--step N (1 through 6)\n";

    print "                 wsdshell.pl is simply a wrapper to call other programs\n";

    print "                 in a sequence, to process the SENSEVAL2 file. Each program\n";

    print "                 this script invokes is considered a step. There are 6\n";

    print "                 such programs called in sequence. The user can directly\n";

    print "                 jump to any desired step using this option. This saves\n";

    print "                 time of certain preprocessing if only some of the programs\n";

    print "                 are to be run with a different set of parameters. The\n";

    print "                 six steps are:\n"; 

    print "                 1. preprocess.pl\n";

    print "                 2. count.pl\n";

    print "                 3. statistic.pl\n";

    print "                 4. nsp2regex.pl\n";

    print "                 5. xml2arff.pl\n";

    print "                 6. Weka\n\n";



    print "--wekaclassifier WEKA_CLASSIFIER\n";

    print "                 Specifies the fully qualified name of the Weka classifier\n";

    print "                 to be used in the final step. The default value used is\n";

    print "                 weka.classifiers.functions.SMO classifier.\n\n";



    print "--verbose        Turns on verbose mode output. Default is off.\n\n";



    print "--prepopts \"OPTIONS\"\n";

    print "                 Specifies any additional options to be passed to \n";

    print "                 preprocess.pl program.\n";

    print "                 Note: All the options must be within double quotes, as\n";

    print "                 would be passed to the program at command line.\n\n";



    print "--countopts \"OPTIONS\"\n";

    print "                 Specifies any additional options to be passed to count.pl\n";

    print "                 program.\n";

    print "                 Note: All the options must be within double quotes, as\n";

    print "                 would be passed to the program at command line.\n\n";



    print "--statopts \"OPTIONS\"\n";

    print "                 Specifies any additional options to be passed to \n";

    print "                 statistic.pl program.\n";

    print "                 Note: All the options must be within double quotes, as\n";

    print "                 would be passed to the program at command line.\n\n";



    print "--wekaopts \"OPTIONS\"\n";

    print "                 Specifies any additional options to be passed while \n";

    print "                 invoking the Weka classifier.\n";

    print "                 Note: All the options must be within double quotes, as\n";

    print "                 would be passed to the program at command line.\n\n";



    print "--help or --?    Prints out this help message.\n\n";



    exit(0);

}



#####################

# getCmdLineOptions #

#####################



# This function parses the command line of the program using the GetOpt::Long package. Additionally, it does some validations on the command line and does some pre-processing tasks like opening input and output files if required.



sub getCmdLineOptions() {



    # Get the state of various options at command line

    # GetOpt::Long leaves anything that it cannot parse in the @ARGV array

    GetOptions(verbose => \$verbose,

               trace => \$trace,

               'token=s' => \$tokenfile,

               'nontoken=s' => \$nontokenfile,

               'stoplist=s' => \$stopfile,

               'wekaclassifier=s' => \$wekaclassifier,

               'wekaopts=s' => \$wekaopts,

               'prepopts=s' => \$prepopts,

               'countopts=s' => \$countopts,

               'statopts=s' => \$statopts,

               'statlib=s' => \$statlibrary,

               'ngram=i' => \$ngrams,

               'step=i' => \$step,

               'help|?' => \$help);



    if ($help == 1) {



        showHelp();

    }



    # Get the name of the input data file, if specified

    if (defined($ARGV[0])) {



        $inputsval2file = shift @ARGV;

        open ($hin, "$inputsval2file") or die "Failed to open file $inputsval2file.\n";

    } else {



        fatalError(1, "Input SENSEVAL2 file not provided.");

    }



    # Get the name of the output file where the user wants to redirect Weka output, if specified

    if (defined($ARGV[0])) {



        $wekaoutputfile = shift @ARGV;

    } # Else, by default the output is shown at STDOUT

}



##############

# fatalError #

##############



# This function should be called when a critical error has occurred and the program cannot proceed anymore. It prints out the specified message at STDERR and exits from the program.



# The first parameter to it indicates whether or not the caller wants this function to print out the program usage.



# Any subsequent parameters are treated as message to be printed. In case of just one additional parameter, we assume that a simple print is required by the caller. In case of multiple subsequent arguments, we assume that a formatted printing is required and we invoke printf



# A newline is added at the end of the message by default

sub fatalError {



    # Check if showUsage should be printed after the error or not

    my $show_usage = shift;



    # Print out the message depending upon the number of additional parameters

    print STDERR "ERROR: ";



    if (@_ > 1) {

        printf STDERR @_;

    } else {

        print STDERR $_[0];

    }



    print STDERR "\n";



    # Print out usage if required

    if ($show_usage) {

        showUsage();

    }



    # Terminate the program, since this is a critical error

    exit(-1);

}



############

# tr_print #

############



# This function prints out the specified trace message at STDERR only when the tracing mode is on



# A single argument is treated as a plain string to be printed using print.

# Multiple arguments are treated as a request for formatted print, with the first argument being the format string.



# A newline is added at the end of the message by default

sub tr_print {



    if ($trace == 1) {



        if (@_ > 1) {

            printf STDERR @_;

        } else {

            print STDERR $_[0];

        }



        print STDERR "\n";

    }

}



############

# vb_print #

############



# This function prints out the specified verbose message at STDERR only when the verbose mode is on



# A single argument is treated as a plain string to be printed using print.

# Multiple arguments are treated as a request for formatted print, with the first argument being the format string.



# A newline is added at the end of the message by default

sub vb_print {



    if ($verbose == 1) {



        if (@_ > 1) {

            printf STDERR @_;

        } else {

            print STDERR $_[0];

        }



        print STDERR "\n";

    }

}



############

# sysCall #

############



# This function simply makes a system function call. It adds an error checking functionality and exits from the program if the system call has returned a non-zero exit status for the command that was executed.



sub sysCall {



    tr_print("About to execute: @_");



    my $status = system(@_);



    # Error handling below taken from documentation of system call from 'perldoc perlfunc'

    unless ($status == 0) {



        if ($? == -1) {



            fatalError(0, "Failed to execute \"@_\": $!");



        } elsif ($? & 127) {



            fatalError(0, "Failed while executing \"@_\". Process died with signal %d, %s coredump",

                     ($? & 127),  ($? & 128) ? 'with' : 'without');

        } else {



            fatalError(0, "Failed while executing \"@_\". Process exited with value %d", $? >> 8);

        }

    }

}



#######################

# MAIN PROGRAM BEGINS #

#######################



# Process the command line options

getCmdLineOptions();



# Check if the user has requested to skip some steps and jump ahead

if ($step > 0 && $step < 7) {

    goto "STEP$step";

}



STEP1:



vb_print("STEP 1: preprocess.pl");



# Initially, we preprocess the input SVAL2 file to obtain a .count file and a .xml file



# Create an array for preprocessing options

my @preprocess_options;



# If user has specified any other preprocessing options, use them

# We do this first, and then specify options that we want, so that our options override the user options where required

if (defined $prepopts) {

    my @prepopts_array = split / /, $prepopts;

    push (@preprocess_options, @prepopts_array);

}



# Specify the token file if known

if (defined $tokenfile) {



    push (@preprocess_options, "--token", $tokenfile);

}



# Specify the non-token file if known

if (defined $nontokenfile) {



    push (@preprocess_options, "--nontoken", $nontokenfile);

}



# Give explicit names to output files

push (@preprocess_options, "--xml", "$inputsval2file.xml");

push (@preprocess_options, "--count", "$inputsval2file.count");





# Finally, specify the name of the input file

push (@preprocess_options, $inputsval2file);



# Now invoke preprocess.pl with all the constructed options

sysCall("preprocess.pl", @preprocess_options);





STEP2:



vb_print("STEP 2: count.pl");



# Now we need to run count.pl program



# Prepare options for count.pl

my @count_options;



# If user has specified any other options, use them

# We do this first, and then specify options that we want, so that our options override the user options where required

if (defined $countopts) {

    my @countopts_array = split / /, $countopts;

    push (@count_options, @countopts_array);

}



# Specify the token file if known

if (defined $tokenfile) {



    push (@count_options, "--token", $tokenfile);

}



# Specify the non-token file if known

if (defined $nontokenfile) {



    push (@count_options, "--nontoken", $nontokenfile);

}



# Specify the stop words file if known

if (defined $stopfile) {



    push (@count_options, "--stop", $stopfile);

}



# Specify the ngram value

push (@count_options, "--ngram", $ngrams);



# Now specify the output file name

push (@count_options, "$inputsval2file.cnt");



# Now specify the input file name

push (@count_options, "$inputsval2file.count");



# Call count.pl with the constructed options

sysCall("count.pl", @count_options);





STEP3:



if ($ngrams == 1) {



    vb_print("Skipping statistic.pl for unigrams");

    goto STEP4;

}



vb_print("STEP 3: statistic.pl");



# Now we call statistic.pl program



# Prepare the options for statistic.pl

my @stat_options;



# If user has specified any other options, use them

# We do this first, and then specify options that we want, so that our options override the user options where required

if (defined $statopts) {

    my @statopts_array = split / /, $statopts;

    push (@stat_options, @statopts_array);

}



# Specify the ngram value

push (@stat_options, "--ngram", $ngrams);



# Now specify the library to be used. Default is log likelyhood

push (@stat_options, $statlibrary);



# Now specify the output file

push (@stat_options, "$inputsval2file.stat");



# Now specify the input file

push (@stat_options, "$inputsval2file.cnt");



# Now give a call to statistic.pl with the constructed options

sysCall("statistic.pl", @stat_options);





STEP4:



vb_print("STEP 4: nsp2regex.pl");



# Now we convert the output of statistic.pl to a set of regular expressions using nsp2regex.pl



# Prepare nsp2regex options

my @nsp2regex_options;



if (defined $tokenfile) {



    push (@nsp2regex_options, "--token", $tokenfile);

}



# Specify the input file (depending upon NGRAM value) and redirect the output to the file we want

if ($ngrams > 1) {

    push (@nsp2regex_options, "$inputsval2file.stat", "> $inputsval2file.regex");

} else {

    push (@nsp2regex_options, "$inputsval2file.cnt", "> $inputsval2file.regex");

}



# Call nsp2regex with the constructed options

sysCall("nsp2regex.pl @nsp2regex_options");



STEP5:



vb_print("STEP 5: xml2arff.pl");



# Now we have to convert the XML file output by count.pl to an ARFF format file which is accepted by Weka. We use the program xml2arff.pl for this

# Redirect the output to an ARFF file that we want

sysCall("xml2arff.pl", "$inputsval2file.regex", "$inputsval2file.xml");





# inserted by Harri Saarikoski 10/05

# STEP 5 creates an ARFF but it is not ready for Weka

# since it contains "%" characters, need to do tilde.pl



vb_print("STEP 6: Tilde replacement");

sysCall("tilde.pl $inputsval2file.xml.arff >$inputsval2file.xml.tilde.arff");





# back again to original code by Ted Pedersen et al.

STEP6:

vb_print("STEP 7: Weka model production");



# We invoke the Weka classifier to run on the output arff file created by xml2arff.pl program

if (defined $wekaoutputfile) {

    sysCall("java $wekaclassifier $wekaopts -t $inputsval2file.xml.tilde.arff -d $wekaoutputfile > weka.result.buffer.txt");

} else {

    sysCall("java $wekaclassifier $wekaopts -t $inputsval2file.xml.tilde.arff");

}



# classification stage inserted by HS

#STEP 8: 

vb_print("STEP 8: WekaClassify");



sysCall("java WekaClassify $wekaopts -l $wekaoutputfile -T $inputsval2file.xml.tilde.arff > $wekaoutputfile.classified.txt");





# one more to run: winnertakesall.pl

# option with ensemblebydist.pl could be provided



#STEP 9:

vb_print("STEP 9: Sense decision (winner takes all)");



sysCall("winnertakeall.pl $wekaoutputfile.classified.txt > $wekaoutputfile.decided.txt");

